;******************************************************************************
;* AVRmultOS cooperative Multitasking Funktionen
;* @author: (c) Markus STRUTZENBERGER struberg [a_t} yahoo.de
;* @date: 2005-09-01
;* @target:Atmel AVR ATtiny26(L)
;*
;* GPL:
;* This program is free software; you can redistribute it and/or modify
;* it under the terms of the GNU General Public License as published by
;* the Free Software Foundation; either version 2 of the License, or
;* (at your option) any later version.
;*
;* This program is distributed in the hope that it will be useful,
;* but WITHOUT ANY WARRANTY; without even the implied warranty of
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;* GNU General Public License for more details.
;*
;* You should have received a copy of the GNU General Public License
;* along with this program; if not, write to the Free Software
;* Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
;*
;******************************************************************************
;*
;* Nomenklatur:
;* Großgeschriebene Labels markieren Funktionen die mit CALL aufgerufen werden müssen.
;* Kleingeschriebene Labels markieren Sprungmarken für RJMP
;* 
;* 8-bit Parameter werden in R16 übergeben,
;* 16-bit Parameter werden in R26,R27 übergeben
;*
;******************************************************************************
;*
;* Funktionsbeschreibung:
;* Diese Datei enthält alle für einen kleinen Multitasking kernel notwendigen
;* Definitionen und Funktionen.
;* Der dafür benötigte Speicherbedarf beträgt 4Byte + 4Byte pro Task, also für
;* 4 Tasks in Summe 20Byte, für 7 Tasks in Summe 32Bytes. Um die Anzahl der 
;* maximal möglichen Tasks zu ändern müssen die .EQU für sTaskX (SPL, SPH, 
;* SleepL und SleepH) entsprechend dupliziert werden.
;* 
;* Die Verteilung der CPU auf die registrierten Tasks wird mittels eines 
;* cooperativen Multitaskings realisiert. Der gerade aktive Task muß also die
;* Kontrolle durch Aufruf der Funktion 
;*   RCALL NextTask
;* manuell abgeben. Die Funktion NextTask holt sich die aufrufende Adresse vom 
;* Stack und speichert sie in der Taskliste ab. Danach wird die in der Taskliste
;* gespeicherte Rücksprungadresse des nächsten Tasks auf den Stack gepushed und
;* ein stinknormales RET durchgeführt, wodurch der nächste Task weitergeführt wird.
;* 
;* Bei herkömmlichen Multitasking kernels geht durch das Sichern 
;* der Register auf den Stack sehr viel SRAM und CPU-Zeit verloren (32-PUSH und 
;* 32 POP Befehle + 32 Byte _PRO_TASK_ am Stack! Um diesen Nachteil auszumerzen 
;* sind die einzelnen Tasks verpflichtet die etwaig von anderen Tasks 
;* benötigten Register SELBST vor Aufruf von MosNextTask oder MosSleep in das 
;* SRAM zu schreiben!
;* Als Alternative müßte PRO TASK ein eigener Stack existieren, was für derartig
;* enge SRAM Ressourcen wie bei den ATTinys nicht sinnvoll erscheint.
;*
;* Für die Funktionen des Multitaskers selbst werden keine Register benötigt!
;*
;* @TODO: Eventuell kann man mit Macros Teile der Funktionalität bei 
;* Nichtverwendung 'ausblenden', um so zu kleinerem Code und schnelleren
;* Taskwechseln zu kommen. zB: 
;* .) wenn kein Sleep benötigt wird
;* .) wenn nur 8bit Sleep benötigt wird
;* .) wenn keine disable task (Sleep=0xffff) + notify benötigt wird
;******************************************************************************

; Definition der Architektur
.NOLIST
;.INCLUDE "TN26def.inc"
.DEVICE ATtiny26
.LIST


;******************************************************************************
;* Speicherbereich für Multitasking
;* Für noch mehr Tasks einfach die xxPC und xxSleep Zeilen duplizieren.
;* Inhalt der Sleep Variablen: 
;* 0x0000 falls der Thread immer aktiv bleiben soll
;* 0xffff falls der Thread blockiert bleiben soll bis ihn jemand anstößt
;* jeder andere Wert hält den Thread für diese Anzahl RT-ticks an und wird
;* sukzessive auf 0 runtergezählt.
;* Mit der Adresse sTimer steht ein 8bit Zähler zur Verfügung, der bei jedem
;* Mos-Tick hochgezählt wird, und für Tastenentprellen, timeouts, etc genutzt
;* werden kann.
;*
;******************************************************************************
; @TODO ev mit Macros hübscher gestalten...
.EQU sTimer       = 0x0060 ; Wird beim Timer0Ovf hochgezählt
.EQU sActiveTask  = 0x0061 ; welcher ist der aktive Task von 0 bis sAnzTasks
.EQU sAnzTasks    = 0x0062 ; wieviele Tasks sind derzeit aktiv? 0..(sAnzTasks-1)
.EQU sTmpL        = 0x0063 ; SRAM temp für Low, da man die nicht pushen kann
.EQU sTmpH        = 0x0064 ; SRAM temp für High, da man die nicht pushen kann
.EQU sTask1PCL    = 0x0065 ; Speicherplatz für PC des Tasks 1
.EQU sTask1PCH    = 0x0066 ; Speicherplatz für PC des Tasks 1
.EQU sTask1SleepL = 0x0067 ; Sleep Count für Task 1
.EQU sTask1SleepH = 0x0068 ; Sleep Count für Task 1
.EQU sTask2PCL    = 0x0069 ; Speicherplatz für PC des Tasks 2
.EQU sTask2PCH    = 0x006a ; Speicherplatz für PC des Tasks 2
.EQU sTask2SleepL = 0x006b ; Sleep Count für Task 2
.EQU sTask2SleepH = 0x006c ; Sleep Count für Task 2
.EQU sTask3PCL    = 0x006d ; Speicherplatz für PC des Tasks 3
.EQU sTask3PCH    = 0x006e ; Speicherplatz für PC des Tasks 3
.EQU sTask3SleepL = 0x006f ; Sleep Count für Task 3
.EQU sTask3SleepH = 0x0070 ; Sleep Count für Task 3
.EQU sTask4PCL    = 0x0071 ; Speicherplatz für PC des Tasks 4
.EQU sTask4PCH    = 0x0072 ; Speicherplatz für PC des Tasks 4
.EQU sTask4SleepL = 0x0073 ; Sleep Count für Task 4
.EQU sTask4SleepH = 0x0074 ; Sleep Count für Task 4
; nächster freier Platz ist 0x0075
;******************************************************************************


;******************************************************************************
;* Intern verwendete Register. 
;* Die Inhalte werden automatisch am Stack zwischengespeichert und können also
;* auch von Userprogrammen verwendet werden.
;******************************************************************************
.DEF rMosParam8   = R16
.DEF rMosParam16L = R26
.DEF rMosParam16H = R27
.DEF rMosTmp      = R18
.DEF rMosAnzT     = R19
.DEF rMosActT     = R20
.DEF rMosTmp2     = R21
;******************************************************************************

;******************************************************************************
;* Hier beginnt der Programmcode des Mos
;******************************************************************************

.CSEG


;******************************************************************************
;* Initialisiere das Multitasking OS
;******************************************************************************
MosInit:
  push rMosTmp
  
  ; Lösche erstmal alle wichtigen Variablen
  clr rMosTmp
  sts sTimer, rMosTmp
  sts sActiveTask, rMosTmp
  sts sAnzTasks, rMosTmp
  
  pop rMosTmp
  ret
;******************************************************************************


;******************************************************************************
;* Füge einen neuen Task in die Taskliste ein
;* @param R26,R27 Adresse des zu startenden Tasks. Diese Adresse wird in der 
;* letzten Stelle der Taskliste eingetragen und der sAnzTasks um 1 erhöht.
;******************************************************************************
MosAddTask:
  push rMosTmp
  push ZH
  push ZL

  lds rMosTmp, sAnzTasks
  lsl rMosTmp       ; *2
  lsl rMosTmp       ; *4

  ldi ZH, HIGH(sTask1PCL)
  ldi ZL, LOW(sTask1PCL)
  add ZL, rMosTmp
  brcc addt_nocarry   ; nur wenn carry gesetzt ist ZH erhöhen
  inc ZH
addt_nocarry:

  ST Z+, rMosParam16L ; *Z++ = R26;
  ST Z+, rMosParam16H ; *Z = R27;
  clr rMosParam16L
  ST Z+, rMosParam16L ; SleepL auf 0
  ST Z, rMosParam16L  ; SleepH auf 0

  lds rMosTmp, sAnzTasks
  inc rMosTmp,
  sts sAnzTasks, rMosTmp

  pop ZL
  pop ZH
  pop rMosTmp
  ret
;******************************************************************************


;******************************************************************************
;* Startet die Scheduler Loop
;* Wie am kleingeschriebenen Namen schon erkennbar handelt es sich hier nicht 
;* um eine Funktion die mit call aufgerufen wird, sondern um einen nicht
;* wiederkehrenden Block der mit JMP (RJMP) aufgerufen werden sollte und
;* die CPU-Zuteilung in einer Endlosschleife durchführt.
;*
;* Typischerweise wird diese Funktion am Ende der reset-Funktion angesprungen,
;* nachdem alle notwendigen Tasks mit AddTask registriert wurden.
;* 
;* @param Diese Funktion benötigt keine Aufrufparameter
;******************************************************************************
mosStartScheduler:
  clr rMosTmp
  sts sActiveTask, rMosTmp    ; Active Task auf 0 initialisieren

  ; timer initialisieren und starten
  ldi R16, 0x02               ; clock/8 prescaler fuer Timer0
  out TCCR0, R16              ; Timer0 Control Register

  in R16, TIMSK
  ori R16, 0x02               ; enable Timer0 overflow Interrupt
  out TIMSK, R16
  sei
  
  ; eigentlich brauchen wir nur den ersten Task anzustoßen, der Rest
  ; erfolgt automatisch durch die Aufrufe von NextTask().
  ; 
  lds ZH, sTask1PCH
  lds ZL, sTask1PCL

  ijmp  ; und auf gehts!
;******************************************************************************


;******************************************************************************
;* Gib die Kontrolle über die CPU an den nächsten aktiven Task ab.
;* Einen aktiven Task erkennt man daran, daß SleepH und SleepL 0 sind.
;* Der Aufruf des nächsten Tasks erfolgt per RET und vorhergehender
;* Verbiegung des Stack-Inhaltes.
;******************************************************************************
MosNextTask:
  sts sTmpL,R19   ; popen geht ja nicht, also ab ins SRAM damit
  sts sTmpH,R20   ; popen geht ja nicht, also ab ins SRAM damit
  pop R20           ; Die Rücksprungadresse des aufrufenden Tasks vom Stack holen und
  pop R19           ; zwischenspeichern

  push ZL           ; Z-Register wegsichern
  push ZH
  push rMosTmp      ; Tmp-Register wegsichern

  ; zuerst muß der PC des Aufrufers in der Taskliste abgelegt werden
  lds rMosTmp, sActiveTask
  lsl rMosTmp       ; *2
  lsl rMosTmp       ; *4

  ldi ZH, HIGH(sTask1PCL)
  ldi ZL, LOW(sTask1PCL)
  add ZL, rMosTmp
  brcc nextt_nocarry1   ; nur wenn carry gesetzt ist ZH erhöhen
  inc ZH
nextt_nocarry1:
  ; *Z zeigt also auf den Task PC
  ST Z, R19        ; *Z = R19;
  STD Z+1, R20         ; *(Z+1) = R20
  ; ENDE Aufrufer PC in Tasklist ablegen

  ; nun muß der nächste aktive Task ermittelt werden
  lds rMosAnzT, sAnzTasks
  lds rMosActT, sActiveTask
nextt_checknextactive:
  inc rMosActT      ; auf den nächsten aktiven Task weiterschalten
  cp rMosAnzT, rMosActT ; ist der Taskzähler am Ende angelangt?
  breq nextt_lasttask   
  adiw ZL,4         ; *Z auf den nächsten Task weiterrücken 
  rjmp nextt_checksleep
nextt_lasttask:
  clr rMosActT      ; wenn das der letzte Task war, dann beginne wieder mit null
  ldi ZH, HIGH(sTask1PCL)
  ldi ZL, LOW(sTask1PCL)

nextt_checksleep:   ; Überprüfe ob der sleep value==0, also der Task aktiv ist
  ldd rMosTmp,Z+2   
  tst rMosTmp
  brne nextt_checknextactive
  ldd rMosTmp, Z+3
  tst rMosTmp
  brne nextt_checknextactive

nextt_isactive:
  sts sActiveTask, rMosActT

  ; der PC des nächsten aktiven Tasks wird in R26, R27 abgelegt
  ld R19, Z
  ldd R20, Z+1

  pop rMosTmp
  pop ZH
  pop ZL
  push R19          ; Die Adresse des nächsten Tasks auf den Stack knallen
  push R20
  lds R19, sTmpL    ; hole den zwischengespeicherten Registerinhalt aus dem SRAM
  lds R20, sTmpH    ; hole den zwischengespeicherten Registerinhalt aus dem SRAM
  ret               ; und ciao bello
;******************************************************************************


;******************************************************************************
;* Halte den aktuellen Task für die in R26/R27 angegebene Anzahl ticks an
;* Enthält R27 0xFF, dann wird der Task solange angehalten, bis er extern 
;* getriggert wird.
;******************************************************************************
MosSleep:
  sts sTmpL, ZL
  sts sTmpH, ZH
  push rMosTmp

  lds rMosTmp, sActiveTask
  lsl rMosTmp       ; *2
  lsl rMosTmp       ; *4

  ldi ZH, HIGH(sTask1PCL)
  ldi ZL, LOW(sTask1PCL)
  add ZL, rMosTmp
  brcc sleep_nocarry1   ; nur wenn carry gesetzt ist ZH erhöhen
  inc ZH
sleep_nocarry1:
  ; *Z zeigt also auf den Task PC
  STD Z+2, R26        ; *(Z+2) = R26;
  STD Z+3, R27        ; *(Z +3)= R27;
  
  pop rMosTmp
  lds ZL, sTmpL
  lds ZH, sTmpH
  rjmp MosNextTask    ; ACHTUNG KEIN CALL! wir brauchen ja die Rücksprungadresse noch
;******************************************************************************

;******************************************************************************
;* Setzt den per rMosParam8 indizierten Thread auf den Status 'running', indem
;* der entsprechende Sleep value auf 0x0000 gesetzt wird.
;******************************************************************************
MosNotify:
  push ZH
  push ZL
  push rMosTmp

  lsl rMosTmp       ; *2
  lsl rMosTmp       ; *4

  ldi ZH, HIGH(sTask1PCL)
  ldi ZL, LOW(sTask1PCL)
  add ZL, rMosTmp
  brcc notify_nocarry   ; nur wenn carry gesetzt ist ZH erhöhen
  inc ZH
notify_nocarry:
  ; *Z zeigt also auf den Task PC
  clr rMosTmp
  STD Z+2, rMosTmp      ; *(Z+2) = 0;
  STD Z+3, rMosTmp      ; *(Z+3) = 0;

  pop rMosTmp
  pop ZL
  pop ZH
  ret
;******************************************************************************


;******************************************************************************
;* Service Routine für Timer 0
;* Zählt gegebenenfalls die Sleep timer der einzelnen Tasks runter.
;* Spezialfälle:
;* Sleep=0x0000  => der task ist aktiv, also so lassen wie es ist
;* Sleep=0xffff  => der task ist disabled, 
;******************************************************************************
MosTimer0ISR:
  push rMosTmp        ; rMosTmp selbst pushen
  in rMosTmp, SREG    ; SREG muß zwischengespeichert werden!
  push rMosTmp        ; nun SREG Wert pushen
  push rMosActT
  push ZL
  push ZH
  push rMosTmp2

  wdr                 ; resete den Watchdog. @TODO wdr hier effizient?

  ; Zähle Timer hoch
  lds rMosTmp, sTimer
  inc rMosTmp,
  sts sTimer, rMosTmp

  ldi ZH, high(sTask1SleepL)
  ldi ZL, low(sTask1SleepL)   ; Z auf den ersten Sleep value stellen
  lds rMosActT, sAnzTasks     ; und zähle auf 0 runter

tisr_checkStart:
  ldd rMosTmp, Z+0    ; lade den low Teil des aktuellen Sleep values
  ldd rMosTmp2, Z+1   ; lade den high Teil des aktuellen Sleep values

  ; prüfe, ob der Task schon aktiv ist
tisr_checkIsActive:
  tst rMosTmp
  brne tisr_checkIsInactive
  tst rMosTmp2
  brne tisr_checkIsInactive
  rjmp tisr_nexttask

tisr_checkIsInactive:
  cpi rMosTmp, 0xff
  brne tisr_countDown
  cpi rMosTmp2, 0xff
  brne tisr_countDown
  rjmp tisr_nexttask
  
tisr_countDown:
  dec rMosTmp
  cpi rMosTmp, 0xff
  brne tisr_storeSleepValue   ; equal=zero, also rMosTmp2 nur runterzählen wenn rMosTmp==0xff
  dec rMosTmp2
tisr_storeSleepValue:
  std Z+0, rMosTmp
  std Z+1, rMosTmp2

tisr_nexttask:
  adiw ZL, 4
  dec rMosActT
  brne tisr_checkStart

tisr_checkEnd: 
  pop rMosTmp2     
  pop ZH
  pop ZL
  pop rMosActT  
  pop rMosTmp          ; SREG popen
  out SREG, rMosTmp     
  pop rMosTmp          ; rMosTmp popen
  reti
;******************************************************************************
